/* Mixmaster version 3  --  (C) 1999 Anonymizer Inc.

   Mixmaster may be redistributed and modified under certain conditions.
   This software is distributed on an "AS IS" basis, WITHOUT WARRANTY OF
   ANY KIND, either express or implied. See the file COPYRIGHT for
   details.

   Menu-based user interface - utility functions
   $Id: menuutil.c,v 1.25 1999/05/14 18:44:29 um Exp $ */


#include "menu.h"
#include <stdarg.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>

int menu_initialized = 0;

#ifdef USE_NCURSES
void cl(int y, int x)
{
  move(y, x);
  hline(' ', COLS - x);
}
#endif

void addprintable(BUFFER *out, int c)
{
  if (c == '\n')
    buf_appendc(out, (char) c);
  else if (c == '\t')
    buf_appends(out, "        ");
  else if (c == '\014')
    buf_appends(out, "^L");
  else if (c == '\r') ;
  else if (c <= 31 || (c >= 128 && c <= 128 + 31))
    buf_appendc(out, '?');
  else
    buf_appendc(out, (char) c);
}

void addprintablebuf(BUFFER *out, BUFFER *in)
{
  int c;

  while ((c = buf_getc(in)) != -1)
    addprintable(out, c);
}

int decode_line(BUFFER *line, int encoded)
{
  BUFFER *out;
  unsigned int i;
  int c;

  out = buf_new();
  for (i = 0; line->data[i] != '\0'; i++) {
    if (encoded && line->data[i] == '=') {
      if (isxdigit(line->data[i + 1]) && isxdigit(line->data[i + 2])) {
	c = hex(line->data[i + 1]) * 16 + hex(line->data[i + 2]);
	i += 2;
	addprintable(out, c);
      } else
	i++;
    } else
      addprintable(out, line->data[i]);
  }

  buf_move(line, out);
  buf_free(out);
  return (0);
}

int decode_header(BUFFER *in)
{
  int encoded = 0;
  int c;
  int err = 0;
  int last = 0;
  BUFFER *out;

  out = buf_new();
  for (in->ptr = 0; in->data[in->ptr] != '\0'; in->ptr++) {
    if (encoded == 'q' && in->data[in->ptr] == '=' &&
	isxdigit(in->data[in->ptr + 1])
	&& isxdigit(in->data[in->ptr + 2])) {
      c = hex(in->data[in->ptr + 1]) * 16 + hex(in->data[in->ptr + 2]);
      in->ptr += 2;
      addprintable(out, c);
    } else if (encoded == 'q' && in->data[in->ptr] == '_')
      buf_appendc(out, ' ');
    else if (in->data[in->ptr] == '=' && in->data[in->ptr + 1] == '?' &&
	     in->data[in->ptr + 2] != '\0') {
      if (last > 0 && out->length > last) {
	out->data[last] = '\0';
	out->length = last;
      }
      in->ptr++;
      while (in->data[++in->ptr] != '?')
	if (in->data[in->ptr] == 0) {
	  err = -1;
	  goto end;
	}
      if (in->data[in->ptr + 1] != '\0' && in->data[in->ptr + 2] == '?') {
	encoded = tolower(in->data[in->ptr + 1]);
	in->ptr += 2;
	if (encoded == 'b') {
	  BUFFER *tmp;

	  tmp = buf_new();
	  decode(in, tmp);
	  addprintablebuf(out, tmp);
	  last = out->length;
	  buf_free(tmp);
	} else if (encoded != 'q')
	  err = 1;
      } else {
	err = -1;
	goto end;
      }
    } else if (encoded && in->data[in->ptr] == '?' &&
	       in->data[in->ptr + 1] == '=') {
      in->ptr++;
      last = out->length;
      encoded = 0;
    } else {
      addprintable(out, in->data[in->ptr]);
      if (!encoded || !isspace(in->data[in->ptr]))
	last = out->length;
    }
  }
end:
  if (err == -1)
    buf_set(out, in);

  buf_move(in, out);
  buf_free(out);
  return (err);
}

int encoding(int encoded, BUFFER *field, BUFFER *content)
{
  if (field->length && content->length &&
      strieq(field->data, "content-transfer-encoding") &&
      strieq(content->data, "quoted-printable"))
    return ('q');
  else
    return encoded;
}

int txtdecode(BUFFER *out, BUFFER *in)
{
  BUFFER *line;

  line = buf_new();
  while (buf_getline(in, line) != -1) {
    buf_nl(line);
    decode_line(line, 1);
    buf_cat(out, line);
  }
  buf_free(line);
  return (0);
}

void menu_init(void)
{
#ifdef USE_NCURSES
  initscr();
  cbreak();
  noecho();
  nonl();
  intrflush(stdscr, FALSE);
  keypad(stdscr, TRUE);
  menu_initialized = 1;
#endif
}

void menu_exit(void)
{
#ifdef USE_NCURSES
  endwin();
#endif
}

#ifdef USE_NCURSES
void askfilename(char *path)
{
  char line[PATHMAX];

  printw("\rFile name: ");
  echo();
  wgetnstr(stdscr, path, PATHMAX);
  noecho();
  printw("\r");
  if (path[0] == '~') {
    char *h;

    if ((h = getenv("HOME")) != NULL) {
      strncpy(line, h, PATHMAX);
      strncat(line, "/", PATHMAX);
      strncat(line, path + 1, PATHMAX);
      strncpy(path, line, PATHMAX);
    }
  }
}

void savemsg(BUFFER *message)
{
  char savename[PATHMAX];
  FILE *f;

  askfilename(savename);
  f = fopen(savename, "a");
  if (f != NULL) {
    buf_write(message, f);
    fclose(f);
  }
}

#endif

int menu_getuserpass(BUFFER *b, int mode)
{
#ifdef USE_NCURSES
  char p[LINELEN];

  if (menu_initialized) {
    cl(LINES - 1, 10);
    if (mode == 0)
      printw("enter passphrase: ");
    else
      printw("re-enter passphrase: ");
    wgetnstr(stdscr, p, LINELEN);
    cl(LINES - 1, 10);
    refresh();
    if (mode == 0)
      buf_appends(b, p);
    else
      return (streq(b->data, p));
    return (0);
  }
#endif
  return (-1);
}
